"""
Integration API for :mod:`loguru`.

.. seealso::

    See :doc:`/integrations/loguru` usage guide.
"""
from __future__ import annotations

from logot._loguru import LoguruCapturer as LoguruCapturer
