"""
Integration API for :mod:`logging`.

.. seealso::

    See :doc:`/log-capturing` usage guide.
"""
from __future__ import annotations

from logot._logging import LoggingCapturer as LoggingCapturer
